;;; Copyright (c) 2000 Matteo Frigo
;;; Copyright (c) 2000 Biomedin, s.r.l.
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;;

;; $Id: gmeteor-core.scm,v 1.7 2003/09/28 00:00:24 matley Exp $

(define-module (gmeteor-core)
  :use-module (gmeteor))

(export 
 ;;  utility functions
 arithmetic-interpolation

 ;; simplex solver
 simplex-solve

 ;; functional approximation
 make-function-constraint
 approximate-function

 ;; filter design
 make-filter-specification
 design-filter
 band band? band/left band/right

 make-filter)

(use-modules (guile))
	     
(dynamic-call "gmeteor_init_gmeteor" 
	      (dynamic-link (string-append gmeteor-libdir "/libgmeteor.so")))

;;------------------------------------------------------------
;; SECTION I: MISCELLANEOUS FUNCTIONS
;;------------------------------------------------------------
(define small-value 1.0e-8)

(define (iota n)
  (let loop ((count (1- n)) (result '()))
    (if (< count 0) result
        (loop (1- count) (cons count result)))))

(define (sigma n f)
  (let loop ((n n) (a 0))
    (if (= n 0) a (loop (- n 1) (+ (f (- n 1)) a)))))

(define (arithmetic-interpolation n left right)
  (if (= left right)
      (lambda (i) left)
      (let ((k1 (- right left))
	    (k2 (- n 1)))
	(lambda (i) (+ left (* (/ i k2) k1))))))

(define (points-in-interval left right density)
  (inexact->exact (floor (+ 3 (* (- right left) density) small-value))))

(define (map-flatten f l)
  (apply append (map f l)))

(define (map-flatten->vector f l)
  (list->vector (map-flatten f l)))

(define (compose f g)
  (lambda (x) 
    (f (g x))))

(define (negate f)
  (lambda (x)
    (- (f x))))

;;------------------------------------------------------------
;; SECTION II: SIMPLEX SOLVER
;;------------------------------------------------------------
; Minimize c^T * x subject to
;   A x = b
;   x >= 0
;
; NROW, NCOL : size of A
; OBJECTIVE : scheme vector containing cost c
; EQUATIONS : a function (lambda (J CONT) ...) that calls
;             (CONT LHS RHS), where LHS = A_J  (the j'th row of A)
;             and RHS = B_J
; ITERATION-LIMIT: parameter for simplex solver
; RETURN: a function (lambda (CODE VALUE PRIMALS DUALS) ...), where
;         CODE : return code, either 'optimal, 'infeasible, or 'unbounded
;         VALUE : optimal value of c^T x, if any
;         PRIMALS : scheme vector containing optimal value of x, if any
;         DUALS : scheme vector containing optimal value of dual variables
(define (simplex-solve nrow ncol objective equations iteration-limit return)
  (let* ((duals (make-vector nrow))
         (primals (make-vector ncol))
         (objective-value (make-vector 1))
         (code (%simplex-solve  ;; call good old fortran routine
		(cons nrow (cons ncol iteration-limit))
		objective
		(lambda (j)
		  (equations j (lambda (lhs rhs) (cons lhs rhs))))
		objective-value
		(cons primals duals))))
    (return code (vector-ref objective-value 0) primals duals)))

;;------------------------------------------------------------
;; SECTION III: APPROXIMATION OF FUNCTIONS
;;------------------------------------------------------------
;
; The functions in this section deal with the problem of
; approximating an arbitrary function by means of linear combinations
; of basis functions.  Specifically, approximation is accomplished by
; defining constraints of the form
;
;   SUM_j c_j b_j(x_i) + w(x_i) y <= f(x_i)
;
; where c_j are the coefficients to be found, b_j are the basis functions,
; x_i is a suitable set of sample points (the `grid'), w is an arbitrary
; weight function, and f is another arbitrary function.  
;
; The approximation is chosen so as to maximize y.
;
; Example:  let b_0(x) = 1, b_1(x) = x^2, and b_2(x) = x^4 . Approximate
; cos(x) by means of linear combinations of b_j(x) on the interval
; [0..1] .  
;
; To solve this problem, proceed as follows.  We set up two constraints
;
;   SUM_j c_j b_j(x) + y <= cos(x)
;   SUM_j c_j b_j(x) - y >= cos(x)
;
; The absolute value of the parameter `y' (y is negative) is thus the maximum 
; approximation error.  By maximizing y, we minimize the error.
; We now convert the problem into a form acceptable to the solver by
; negating the second constraint:
;
;     SUM_j c_j b_j(x) + y <=   cos(x)
;   - SUM_j c_j b_j(x) + y <= - cos(x)
;
; In Scheme:
;
; (let* ((basis (lambda (j) (if (= j 0) 
; 			      (lambda (x) 1.0)
; 			      (lambda (x) (expt x (* 2 j))))))
;        (w (lambda (x) 1.0)) ;; weight
;        (constraints
;         (list
;          (make-function-constraint 0 1.0 basis w cos)
;          (make-function-constraint 0 1.0 (compose negate basis) w
;                                    (negate cos)))))
;   (approximate-function 3 100 100 constraints (lambda x (display x))))
;
; =>
; (optimal -4.18143628799928e-5 #(0.99995818563712 -0.499241630320456 0.0396275649143552))
;

; Set up a constraint of the form
;   SUM_j c_j b_j(x) + w(x) y <= f(x)
; over an interval.
;
; [LEFT..RIGHT] : interval
; BASIS : (lambda (j) (lambda (x) b_j(x)))
; WEIGHT : the function w
; BOUND : the function f
;
(define (make-function-constraint left right basis weight bound)
  (lambda (grid-density)
    (let* ((nvariables 
	    (points-in-interval left right grid-density))

	   (tabulate 
	    (let ((g (arithmetic-interpolation nvariables left right)))
	      (lambda (f)
		(map (compose f g) (iota nvariables))))))

      (lambda (msg)
	(case msg
	  ((tabulate-basis) 
	   (compose tabulate basis))
	  ((tabulate-bound)
	   (tabulate bound))
	  ((tabulate-weight)
	   (tabulate weight)))))))

;
; Approximate a function given a list of constraints.
;
; NCOEF : number of elements in basis
; GRID-DENSITY : density at which constraints are sampled.  An interval
;         of length X contains (* X GRID-DENSITY) sample points.
; ITERATION-LIMIT : limit for simplex solver
; CONSTRAINTS : list of constraints produced by MAKE-FUNCTION-CONSTRAINT
; RETURN : continuation = (lambda (CODE Y COEF) ...) where
;   CODE : see SIMPLEX-SOLVE
;   Y    : optimal value of Y, if any
;   COEF : scheme vector of the optimal c_j's, if any
;
; Method of solution: dual simplex.  The sampled constraints define
; a linear-programming problem
; 
; max y subject to
;  B c + w y <= f
; 
; We solve the dual problem
;
; min f' x subject to
;  w' x = 1
;  B' x = 0
;  x >= 0
;
; y and c are now the optimal dual variables of the dual problem.
;
(define (approximate-function ncoef grid-density iteration-limit 
			      constraints return)
  (let ((constraints* (map (lambda (s) (s grid-density)) constraints)))
    (let ((objective-function 
	   ; The objective function is f' x
	   (map-flatten->vector (lambda (s) (s 'tabulate-bound))
				constraints*))
	  (equations
	   (lambda (i return)
	     (if (= i 0)
		 ;; weight equation  w' x = 1
		 (return (map-flatten->vector 
			  (lambda (s) (s 'tabulate-weight))
			  constraints*)
			 1.0)
		 ;; equations for the filter coefficients, B' x = 0 
		 (return (map-flatten->vector 
			  (lambda (s) ((s 'tabulate-basis) (- i 1)))
			  constraints*)
			 0.0)))))
      (simplex-solve 
       (+ 1 ncoef)
       (vector-length objective-function)
       objective-function
       equations
       iteration-limit
       (lambda (solution obj primals duals)
	 (case solution
	   ((optimal)
	    (let ((dduals (vector->list duals)))
	      (return solution 
		      (car dduals)
		      (list->vector (cdr dduals)))))
	   ((iteration-limit-reached) (return solution #f #f))
	   ;; swap infeasible<->unbounded since we are solving the dual problem
	   ((infeasible) (return 'unbounded #f #f))
	   ((unbounded) (return 'infeasible #f #f))))))))


;;------------------------------------------------------------
;; SECTION IV: FILTER DESIGN
;;------------------------------------------------------------
(define make-filter-specification 
  (lambda (order sense left right response weight bound)
    (lambda (derivative-of-trig sampling-frequency)
      (let ((check-freq
	     (lambda (freq)
	       (if (not (and (>= freq 0) 
			     (<= freq (/ sampling-frequency 2))))
		   (error "invalid frequency" freq))))
	    (dt (/ 1 sampling-frequency))
	    (trig (derivative-of-trig order))
	    (adjust-for-sense 
	     (lambda (f)
	       (if (eq? sense '<) f (negate f)))))

	(begin
	  (check-freq left)
	  (check-freq right)
	  (lambda (grid-density)
	    ((make-function-constraint 
	      left
	      right
	      (lambda (i)
		(adjust-for-sense (response (trig dt i))))
	      weight
	      (adjust-for-sense bound))
	     (* 2 grid-density dt))))))))

(define design-filter
  (let ()
    (define 2pi (* 8 (atan 1)))
    (define (modf x) (- x (floor x)))
    (define (halve l) (map (lambda (x) (* x 1/2)) l))
    (define (nhalve l) (map (lambda (x) (* x -1/2)) l))

    ;; compute the n-th derivative of cos(2*pi*k*x)
    (define (dcos2pi n dt k)
      (if (= n 0)
	  (lambda (x)
	    (cos (* 2pi (modf (* k dt x)))))
	  (let ((d (dsin2pi (- n 1) dt k)))
	    (lambda (x)
	      (- (* 2pi k dt (d x)))))))

    ;; compute the n-th derivative of sin(2*pi*k*x)
    (define (dsin2pi n dt k)
      (if (= n 0)
	  (lambda (x)
	    (sin (* 2pi (modf (* k dt x)))))
	  (let ((d (dcos2pi (- n 1) dt k)))
	    (lambda (x)
	      (* 2pi k dt (d x))))))

    (define (type-I-filter return) 
      (return (lambda (n)
		(lambda (dt i)
		  (dcos2pi n dt i)))
	      (lambda (filter-length) (/ (+ filter-length 1) 2))
	      (lambda (l)
		(append (halve (reverse (cdr l))) 
			(list (car l))
			(halve (cdr l))))))

    (define (type-II-filter return) 
      (return (lambda (n)
		(lambda (dt i)
		  (dcos2pi n dt (+ i 1/2))))
	      (lambda (filter-length) (/ filter-length 2))
	      (lambda (l)
		(append (halve (reverse l)) 
			(halve l)))))

    (define (type-III-filter return) 
      (return (lambda (n)
		(lambda (dt i)
		  (dsin2pi n dt (+ i 1))))
	      (lambda (filter-length) (/ (- filter-length 1) 2))
	      (lambda (l)
		(append (nhalve (reverse l))
			(list 0.0)
			(halve l)))))
    
    (define (type-IV-filter return) 
      (return (lambda (n)
		(lambda (dt i)
		  (dsin2pi n dt (+ i 1/2))))
	      (lambda (filter-length) (/ filter-length 2))
	      (lambda (l)
		(append (nhalve (reverse l))
			(halve l)))))

    (define (filter-type filter-length symmetry)
      (case symmetry
	((cosine) 
	 (if (odd? filter-length) type-I-filter type-II-filter))
	((sine) 
	 (if (odd? filter-length) type-III-filter type-IV-filter))))

    (lambda (filter-length symmetry sampling-frequency harmonic
			   constraints grid-density iteration-limit
			   return)
      ((filter-type filter-length symmetry)
       (lambda (derivative-of-trig compute-ncoef reconstruct-coef)
	 (let* ((der* (lambda (n)
		       (let ((f (derivative-of-trig n)))
			 (lambda (dt i)
			   (f dt (harmonic i))))))
		(ncoef (compute-ncoef filter-length))
		(specl 
		 (map (lambda (s) (s der* sampling-frequency)) constraints)))
	   (approximate-function
	    ncoef grid-density iteration-limit specl
	    (lambda (solution y coefs)
	      (case solution
		((optimal)
		 (return solution 
			 (list->vector
			  (reconstruct-coef (vector->list coefs)))
			 y
			 (der* 0)
			 coefs))
		(else
		 (return solution #f #f #f #f)))))))))))

;;------------------------------------------------------------
;; SECTION V: FILTER ADT (i.e., the advanced interface)
;;------------------------------------------------------------
(define (ensure truth complaint)
  (if (not truth) (error complaint)))

(define band 'unbound)
(define band? 'unbound)
(let ((band-token (cons #f #f)))
  (set! band
	(lambda (l u)
	  (ensure (real? l) "LEFT frequency must be a real number")
	  (ensure (real? u) "RIGHT frequency must be a real number")
	  (ensure (<= l u) "LEFT <= RIGHT must hold")
	  (list band-token l u)))
  (set! band?
	(lambda (x)
	  (and (pair? x)
	       (eq? (car x) band-token)))))

(define (band/left i) (cadr i))
(define (band/right i) (caddr i))

(define (make-filter)
  (let (;; user-defined parameters
	(filter-length #f)
	(symmetry 'cosine)
	(sampling-frequency 1)
	(specifications '())
	(readable-specifications '())
	(grid-density #f)
	(plot-grid-density #f)
	(plot-absolute #f)
	(harmonic (lambda (i) i))
	(verbose #f)
	(iteration-limit 10000)
	(output-file "-")
	(plot-file #f)
	(title "UNTITLED")

	;; state variables computed by gmeteor
	(coefficients #f)
	(y #f)
	(H #f)
	(result #f)
	(duals #f)
	(trig #f))

    (define (once f)
      (let ((already? #f))
	(lambda (x)
	  (if (not already?)
	      (begin
		(f x)
		(set! already? #t))))))

    (define (set-filter-length! n)
      (ensure (integer? n) "filter length must be an integer")
      (set! filter-length n))
    (define (get-filter-length) filter-length)

    (define (set-symmetry! x)
      (ensure (or (eq? x 'sine) (eq? x 'cosine))
	      "symmetry must be either 'SINE or 'COSINE")
      (set! symmetry x))
    (define (get-symmetry) symmetry)

    (define (set-sampling-frequency! x)
      (ensure (and (real? x) (positive? x))
	      "sampling frequency must be a positive number")
      (set! sampling-frequency x))
    (define (get-sampling-frequency) sampling-frequency)

    (define (set-harmonic! f)
      (ensure (procedure? f) "harmonic must be a procedure of one argument")
      (set! harmonic f))

    (define make-readable-specification
      (let ()
	(define (number-maybe x)
	  (if (number? x) x "<procedure>"))
	(lambda (type l u sense bound weight)
	  (list type sense l u 
		(number-maybe bound)
		(number-maybe weight)))))

    (define really-add-specification! 
      (lambda (order type i sense response bound weight)
	(let ((l (cond ((number? i) i)
		       ((band? i) (band/left i))
		       (else (ensure #f "invalid band"))))
	      (u (cond ((number? i) i)
		       ((band? i) (band/right i))
		       (else (ensure #f "invalid band"))))
	      (tsense (case sense
			((<) '(<))
			((<=) '(<))
			((=) '(< >))
			((>) '(>))
			((>=) '(>))
			(else (ensure #f "unknown SENSE"))))
	      (tbound 
	       ;; BOUND can be either a constant or a procedure
	       (cond ((real? bound) (lambda (x) bound))
		     ((procedure? bound) bound)
		     (else (ensure #f "invalid BOUND"))))
	      (tweight 
	       (cond ((real? weight) (lambda (x) weight))
		     ((procedure? weight) weight)
		     (else (ensure #f "invalid WEIGHT")))))
	  (for-each
	   (lambda (sense)
	     (set! readable-specifications
		   (cons (make-readable-specification 
			  type l u sense bound weight)
			 readable-specifications))
	     (set! specifications
		   (cons (make-filter-specification order sense l u
						    response tweight tbound)
			 specifications)))
	   tsense))))
    
    (define (default-response H) H)

    (define (add-specification! type band sense . args)
      (case type 
	((limit)
	 (apply really-add-specification!
		0
		type
		band
		sense
		default-response
		args))
	((derivative) 
	 (begin
	   (ensure (not (null? args)) "invalid derivative specification")
	   (let ((order (car args)))
	     (apply really-add-specification! 
		    order
		    (list type order)
		    band
		    sense
		    default-response
		    (cdr args)))))
	(else (ensure #f "unknown specification type"))))
    
    (define (set-grid-density! x)
      (ensure (and (integer? x) (positive? x))
	      "grid density must be a positive integer")
      (set! grid-density x))
    (define (get-grid-density) grid-density)

    (define (set-plot-grid-density! x)
      (ensure (and (integer? x) (positive? x))
	      "plot grid density must be a positive integer")
      (set! plot-grid-density x))
    (define (get-plot-grid-density) plot-grid-density)

    (define (set-plot-absolute! x)
      (set! plot-absolute x))
    (define (set-verbose! x)
      (set! verbose x))
    (define (set-iteration-limit! x)
      (ensure (and (integer? x) (positive? x))
	      "iteration limit must be a positive integer")
      (set! iteration-limit x))
    (define (set-output-file! f)
      (ensure (or (not f) (string? f))
	      "output file must be a string")
      (set! output-file f))
    (define (set-plot-file! f)
      (ensure (or (not f) (string? f))
	      "plot output file must be a string")
      (set! plot-file f))
    (define (set-title! x)
      (set! title x))

    (define (get-result) result)
    (define (get-coefficients) coefficients)
    (define (get-y) y)
    (define (get-harmonics) duals)
    (define (get-H) H)

    (define (display-verbose)
      (begin 
	(display ";; ***** ") (display title) (display " *****") (newline)
	(display ";; Filter length: ") (display filter-length) (newline)
	(display ";; Symmetry: ") (display symmetry) (newline)
	(display ";; Sampling frequency: ") (display sampling-frequency)
	(newline)
	(display ";; Grid density: ") (display grid-density)
	(newline)
	(display ";; Deviation parameter: y=") (display y)
	(newline)
	(display ";;") (newline)
	(display ";; Specifications: ") (newline)
	(for-each (lambda (s)
		    (begin (display ";; ") (display s) (newline)))
		  readable-specifications)
	(display ";;") (newline)
	(display ";; Coefficients: ") (newline) (newline)))

    (define (display-coef)
      (case result
	((optimal)
	 (begin
	   (if verbose (display-verbose))
	   (for-each (lambda (x)
		       (begin (display (exact->inexact x))
			      (newline)))
		     (vector->list coefficients))))
	((iteration-limit-reached)
	 (begin
	   (display ";; Reached iteration limit.")
	   (newline)))
	((infeasible)
	 (begin
	   (display ";; The problem is infeasible.")
	   (newline)))
	((unbounded)
	 (begin
	   (display ";; The problem is unbounded.") (newline)
	   (display ";; This situation occurs when all weights are 0,")
	   (newline)
	   (display ";; or when there are not enough constraints on the")
	   (newline)
	   (display ";; frequency response.")
	   (newline)))))

    (define (display-plot)
      (case result
	((optimal)
	 (let ((maybe-abs (if plot-absolute abs (lambda (x) x))))
	   (for-each (lambda (f)
		       (begin (display (exact->inexact f))
			      (display " ")
			      (display (exact->inexact 
					(maybe-abs (H f))))
			      (newline)))
		     (map (arithmetic-interpolation plot-grid-density 
						    0
						    (/ sampling-frequency 2))
			  (iota plot-grid-density)))))))
    
    (define (maybe-with-output-to-file f thunk)
      (if f
	  (if (string=? f "-")
	      (thunk)
	      (with-output-to-file f thunk))))

    (define (gmeteor-display)
      (begin
	(maybe-with-output-to-file output-file display-coef)
	(maybe-with-output-to-file plot-file display-plot)))

    (define (check-parameters)
      (begin
	(ensure (and (integer? filter-length) (positive? filter-length))
		"filter length not specified")
	(if (not grid-density)
	    (set! grid-density (+ 17 (* 10 filter-length))))
	(if (not plot-grid-density)
	    (set! plot-grid-density grid-density))
	(ensure (or (eq? symmetry 'sine) (eq? symmetry 'cosine))
		"invalid symmetry")
	(ensure (and (integer? grid-density) (positive? grid-density))
		"filter length must be a positive integer")
	(ensure (and (integer? iteration-limit) (positive? iteration-limit))
		"iteration limit must be a positive integer")
	(ensure (and (real? sampling-frequency) (positive? sampling-frequency))
		"sampling frequency must be a positive real number")))

    (define (solve!)
      (check-parameters)
      (design-filter filter-length
		     symmetry
		     sampling-frequency
		     harmonic
		     specifications
		     grid-density
		     iteration-limit
		     (lambda (c0 coef0 y0 trig0 duals0)
		       (set! result c0)
		       (set! coefficients coef0)
		       (set! y y0)
		       (set! trig trig0)
		       (set! duals duals0)
		       (if (eq? result 'optimal)
			   (set! H 
				 (let ((dt (/ 1 sampling-frequency))
				       (nduals (vector-length duals)))
				   (lambda (f)
				     (sigma nduals
					    (lambda (i)
					      (* ((trig dt i) f)
						 (vector-ref duals i))))))))
		       )))

    (let ((set-filter-length!* (once set-filter-length!))
	  (set-sampling-frequency!* (once set-sampling-frequency!))
	  (set-grid-density!* (once set-grid-density!))
	  (set-output-file!* (once set-output-file!))
	  (set-plot-file!* (once set-plot-file!))
	  (set-plot-absolute!* (once set-plot-absolute!))
	  (set-plot-grid-density!* (once set-plot-grid-density!))
	  (set-iteration-limit!* (once set-iteration-limit!)))
	  
    (lambda (msg)
      (case msg
	((solve!) solve!)
	((display) gmeteor-display)
	((get-result) get-result)
	((get-coefficients) get-coefficients)
	((get-y) get-y)
	((get-harmonics) get-harmonics)
	((get-H) get-H)
	((set-filter-length!) set-filter-length!)
	((get-filter-length) get-filter-length)
	((set-symmetry!) set-symmetry!)
	((get-symmetry) get-symmetry)
	((set-sampling-frequency!) set-sampling-frequency!)
	((get-sampling-frequency) get-sampling-frequency)
	((add-specification!) add-specification!)
	((set-grid-density!) set-grid-density!)
	((get-grid-density) get-grid-density)
	((set-plot-grid-density!) set-plot-grid-density!)
	((get-plot-grid-density) get-plot-grid-density)
	((set-plot-absolute!) set-plot-absolute!)
	((set-verbose!) set-verbose!)
	((set-iteration-limit!) set-iteration-limit!)
	((set-output-file!) set-output-file!)
	((set-plot-file!) set-plot-file!)
	((set-title!) set-title!)
	((set-harmonic!) set-harmonic!)

	((set-filter-length!*) set-filter-length!*)
	((set-grid-density!*) set-grid-density!*)
	((set-sampling-frequency!*) set-sampling-frequency!*)
	((set-output-file!*) set-output-file!*)
	((set-plot-file!*) set-plot-file!*)
	((set-plot-absolute!*) set-plot-absolute!*)
	((set-plot-grid-density!*) set-plot-grid-density!*)
	((set-iteration-limit!*) set-iteration-limit!*) )))))
